@@ -8,7 +8,7 @@
  *
  *    In some countries OX, OX Open-Xchange, open xchange and OXtender
  *    as well as the corresponding Logos OX Open-Xchange and OX are registered
- *    trademarks of the Open-Xchange, Inc. group of companies.
+ *    trademarks of the OX Software GmbH group of companies.
  *    The use of the Logos is not covered by the GNU General Public License.
  *    Instead, you are allowed to use these Logos according to the terms and
  *    conditions of the Creative Commons License, Version 2.5, Attribution,
@@ -28,7 +28,7 @@
  *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
  *    given Attribution for the derivative code and a license granting use.
  *
- *     Copyright (C) 2004-2014 Open-Xchange, Inc.
+ *     Copyright (C) 2016-2020 OX Software GmbH
  *     Mail: info@open-xchange.com
  *
  *
@@ -49,25 +49,25 @@
 
 package com.openexchange.push.imapidle.osgi;
 
-import org.osgi.framework.BundleContext;
-import org.osgi.framework.ServiceReference;
-import org.osgi.util.tracker.ServiceTrackerCustomizer;
 import com.hazelcast.core.HazelcastInstance;
 import com.openexchange.config.ConfigurationService;
+import com.openexchange.context.ContextService;
 import com.openexchange.database.DatabaseService;
 import com.openexchange.groupware.delete.DeleteListener;
-import com.openexchange.hazelcast.configuration.HazelcastConfigurationService;
 import com.openexchange.mail.service.MailService;
 import com.openexchange.mailaccount.MailAccountDeleteListener;
 import com.openexchange.osgi.HousekeepingActivator;
-import com.openexchange.push.PushManagerService;
+import com.openexchange.push.PushListenerService;
 import com.openexchange.push.imapidle.ImapIdleConfiguration;
 import com.openexchange.push.imapidle.ImapIdleDeleteListener;
 import com.openexchange.push.imapidle.ImapIdleMailAccountDeleteListener;
 import com.openexchange.push.imapidle.ImapIdlePushManagerService;
+import com.openexchange.push.imapidle.locking.ImapIdleClusterLock;
 import com.openexchange.sessiond.SessiondService;
+import com.openexchange.sessionstorage.SessionStorageService;
 import com.openexchange.threadpool.ThreadPoolService;
 import com.openexchange.timer.TimerService;
+import com.openexchange.user.UserService;
 
 
 /**
@@ -87,48 +87,39 @@ public class ImapIdleActivator extends HousekeepingActivator {
 
     @Override
     protected Class<?>[] getNeededServices() {
-        return new Class<?>[] { DatabaseService.class, TimerService.class, MailService.class, ConfigurationService.class, HazelcastConfigurationService.class, SessiondService.class, ThreadPoolService.class };
+        return new Class<?>[] { DatabaseService.class, TimerService.class, MailService.class, ConfigurationService.class, SessiondService.class,
+            ThreadPoolService.class, ContextService.class, UserService.class, PushListenerService.class };
     }
 
     @Override
     protected void startBundle() throws Exception {
-        HazelcastConfigurationService hzConfigService = getService(HazelcastConfigurationService.class);
-
-        final boolean hzEnabled = hzConfigService.isEnabled();
-        if (hzEnabled) {
-            final BundleContext context = this.context;
-            ServiceTrackerCustomizer<HazelcastInstance, HazelcastInstance> stc = new ServiceTrackerCustomizer<HazelcastInstance, HazelcastInstance>() {
-
-                @Override
-                public HazelcastInstance addingService(ServiceReference<HazelcastInstance> reference) {
-                    HazelcastInstance hzInstance = context.getService(reference);
-                    addService(HazelcastInstance.class, hzInstance);
-                    return hzInstance;
-                }
-
-                @Override
-                public void modifiedService(ServiceReference<HazelcastInstance> reference, HazelcastInstance service) {
-                    // Nothing
-                }
-
-                @Override
-                public void removedService(ServiceReference<HazelcastInstance> reference, HazelcastInstance service) {
-                    removeService(HazelcastInstance.class);
-                    context.ungetService(reference);
-                }
-            };
-            track(HazelcastInstance.class, stc);
-            openTrackers();
-        }
-
         ImapIdleConfiguration configuration = new ImapIdleConfiguration();
         configuration.init(this);
-        registerService(PushManagerService.class, ImapIdlePushManagerService.newInstance(configuration, this));
+
+        // Check if Hazelcast-based locking is enabled
+        if (ImapIdleClusterLock.Type.HAZELCAST.equals(configuration.getClusterLock().getType())) {
+            // Start tracking for Hazelcast
+            ImapIdleRegisteringTracker registeringTracker = new ImapIdleRegisteringTracker(true, configuration, this, context);
+            track(registeringTracker.getFilter(), registeringTracker);
+        } else {
+            // Register PushManagerService instance
+            ImapIdleRegisteringTracker registeringTracker = new ImapIdleRegisteringTracker(false, configuration, this, context);
+            track(registeringTracker.getFilter(), registeringTracker);
+            trackService(HazelcastInstance.class);
+        }
+        trackService(SessionStorageService.class);
+        openTrackers();
 
         registerService(MailAccountDeleteListener.class, new ImapIdleMailAccountDeleteListener());
         registerService(DeleteListener.class, new ImapIdleDeleteListener());
     }
 
+    @Override
+    protected void stopBundle() throws Exception {
+        stopPushManagerSafe();
+        super.stopBundle();
+    }
+
     @Override
     public <S> boolean addService(Class<S> clazz, S service) {
         return super.addService(clazz, service);
@@ -139,4 +130,14 @@ public class ImapIdleActivator extends HousekeepingActivator {
         return super.removeService(clazz);
     }
 
+    /**
+     * Stops the push manager.
+     */
+    static void stopPushManagerSafe() {
+        ImapIdlePushManagerService pushManager = ImapIdlePushManagerService.getInstance();
+        if (null != pushManager) {
+            pushManager.stopAllListeners();
+        }
+    }
+
 }
